import re
from typing import Optional, Union, cast

from django.core.exceptions import ImproperlyConfigured
from django.urls import reverse
from typing_extensions import Literal, TypedDict, TypeVar, override

from django_backblaze_b2.options import (
    BackblazeB2StorageOptions,
    CDNConfig,
    PossibleB2StorageOptions,
    ProxiedBucketNames,
)
from django_backblaze_b2.storage import BackblazeB2Storage, logger


class _SdkBucketDict(TypedDict):
    """See https://github.com/Backblaze/b2-sdk-python/blob/2c85182c82ee09b7db7216d70567aafb87f31536/b2sdk/bucket.py#L1148"""  # noqa: E501

    bucketType: Literal["allPublic", "allPrivate"]  # noqa: N815


class PublicStorage(BackblazeB2Storage):
    """
    Storage that requires no authentication to view.
    If the bucket is public, returns the bucket's url, or CDN if configured
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._bucket_type: Optional[Literal["allPublic", "allPrivate"]] = None
        self._cdn_config: Optional[CDNConfig] = kwargs.get("opts", {}).get("cdn_config")

    @override
    def _get_file_url(self, name: str) -> str:
        if not self._is_public_bucket():
            return reverse("django_b2_storage:b2-public", args=[name])
        if self._cdn_config:
            file_url = self.get_backblaze_url(name)
            cdn_url_base = self._cdn_config["base_url"].replace("https://", "").replace("http://", "").strip("/")
            if self._cdn_config["include_bucket_url_segments"]:
                return re.sub(r"f\d+\.backblazeb2\.com", cdn_url_base, file_url)
            return re.sub(r"f\d+\.backblazeb2\.com/file/[^/]+/", cdn_url_base + "/", file_url)
        return self.get_backblaze_url(name)

    def _is_public_bucket(self) -> bool:
        if self._bucket_type is None:
            bucket_dict: _SdkBucketDict = self.bucket.as_dict()
            if bucket_dict.get("bucketType") is None:  # sometimes this happens due to cached values
                logger.debug(f"Re-retrieving bucket info for {bucket_dict}")
                self._refresh_bucket()
                bucket_dict = self.bucket.as_dict()
            self._bucket_type = bucket_dict.get("bucketType")
        return self._bucket_type == "allPublic"

    @override
    def _validate_options(self, options: PossibleB2StorageOptions, from_constructor: bool = False) -> None:
        super()._validate_options(options, from_constructor)
        if from_constructor:
            _validate_proxy_class_options(options)
        if options.get("cdn_config"):
            if not isinstance(options["cdn_config"], dict):
                raise ImproperlyConfigured("django-backblaze-b2 cdn_config must be a dict")
            if not isinstance(options["cdn_config"].get("base_url"), str):
                raise ImproperlyConfigured("cdn_config.base_url must be a string")
            if not isinstance(options["cdn_config"].get("include_bucket_url_segments"), bool):
                logger.debug("will treat cdn_config.include_bucket_url_segments to False")

    @override
    def _extract_constructor_options(
        self, constructor_kwargs: PossibleB2StorageOptions, opts_args: PossibleB2StorageOptions
    ) -> PossibleB2StorageOptions:
        extracted = super()._extract_constructor_options(constructor_kwargs, opts_args)
        return _with_bucket_name_extracted_from_settings(extracted, specific_bucket="public")

    @override
    def _get_options_from_django_settings(self) -> BackblazeB2StorageOptions:
        from_settings = super()._get_options_from_django_settings()
        return _with_bucket_name_extracted_from_settings(from_settings, specific_bucket="public")


class LoggedInStorage(BackblazeB2Storage):
    """Storage that requires authentication to view or download files"""

    @override
    def _get_file_url(self, name: str) -> str:
        return reverse("django_b2_storage:b2-logged-in", args=[name])

    @override
    def _validate_options(self, options: PossibleB2StorageOptions, from_constructor: bool = False) -> None:
        super()._validate_options(options, from_constructor)
        if from_constructor:
            _validate_proxy_class_options(options)

    @override
    def _extract_constructor_options(
        self, constructor_kwargs: PossibleB2StorageOptions, opts_args: PossibleB2StorageOptions
    ) -> PossibleB2StorageOptions:
        extracted = super()._extract_constructor_options(constructor_kwargs, opts_args)
        return _with_bucket_name_extracted_from_settings(extracted, specific_bucket="logged_in")

    @override
    def _get_options_from_django_settings(self) -> BackblazeB2StorageOptions:
        from_settings = super()._get_options_from_django_settings()
        return _with_bucket_name_extracted_from_settings(from_settings, specific_bucket="logged_in")


class StaffStorage(BackblazeB2Storage):
    """Storage that requires staff permission to view or download files"""

    @override
    def _get_file_url(self, name: str) -> str:
        return reverse("django_b2_storage:b2-staff", args=[name])

    @override
    def _validate_options(self, options: PossibleB2StorageOptions, from_constructor: bool = False) -> None:
        super()._validate_options(options, from_constructor)
        if from_constructor:
            _validate_proxy_class_options(options)

    @override
    def _extract_constructor_options(
        self, constructor_kwargs: PossibleB2StorageOptions, opts_args: PossibleB2StorageOptions
    ) -> PossibleB2StorageOptions:
        extracted = super()._extract_constructor_options(constructor_kwargs, opts_args)
        return _with_bucket_name_extracted_from_settings(extracted, specific_bucket="staff")

    @override
    def _get_options_from_django_settings(self) -> BackblazeB2StorageOptions:
        from_settings = super()._get_options_from_django_settings()
        return _with_bucket_name_extracted_from_settings(from_settings, specific_bucket="staff")


def _validate_proxy_class_options(options: PossibleB2StorageOptions) -> None:
    if "bucket" in options:
        raise ImproperlyConfigured("May not specify 'bucket' in proxied storage class")
    if "application_key_id" in options or "application_key" in options or "realm" in options:
        raise ImproperlyConfigured("May not specify auth credentials in proxied storage class")


B2Opts = TypeVar("B2Opts", bound=Union[BackblazeB2StorageOptions, PossibleB2StorageOptions])


def _with_bucket_name_extracted_from_settings(
    options: B2Opts, specific_bucket: Literal["public", "staff", "logged_in"]
) -> B2Opts:
    proxied_storage_bucket_name = (
        cast(PossibleB2StorageOptions, options).get("specific_bucket_names", ProxiedBucketNames()).get(specific_bucket)
    )
    if proxied_storage_bucket_name:
        return cast(B2Opts, {**options, "bucket": options["specific_bucket_names"][specific_bucket]})
    return cast(B2Opts, {**options})



from datetime import datetime
from hashlib import sha3_224 as hash
from logging import getLogger
from typing import IO, Any, Callable, Dict, List, Optional, Tuple, cast

from b2sdk.v2 import AbstractAccountInfo, AuthInfoCache, B2Api, Bucket, InMemoryAccountInfo, SqliteAccountInfo
from b2sdk.v2.exception import FileNotPresent, NonExistentBucket
from django.core.cache.backends.base import BaseCache
from django.core.exceptions import ImproperlyConfigured
from django.core.files.base import File
from django.core.files.storage import Storage
from django.utils.deconstruct import deconstructible
from typing_extensions import NotRequired, TypedDict, TypeVar, Unpack

from django_backblaze_b2.b2_file import B2File
from django_backblaze_b2.cache_account_info import DjangoCacheAccountInfo
from django_backblaze_b2.options import (
    BackblazeB2StorageOptions,
    DjangoCacheAccountInfoConfig,
    PossibleB2StorageOptions,
    SqliteAccountInfoConfig,
    get_default_b2_storage_options,
)

logger = getLogger("django-backblaze-b2")


class _SdkBaseFileInfoDict(TypedDict):
    """See https://github.com/Backblaze/b2-sdk-python/blob/2c85182c82ee09b7db7216d70567aafb87f31536/b2sdk/file_version.py"""  # noqa: E501

    fileId: str  # noqa: N815
    fileName: str  # noqa: N815
    fileInfo: dict  # noqa: N815


class _SdkFileInfoDict(_SdkBaseFileInfoDict, total=False):
    """See https://github.com/Backblaze/b2-sdk-python/blob/2c85182c82ee09b7db7216d70567aafb87f31536/b2sdk/file_version.py#L143"""  # noqa: E501

    size: int
    uploadTimestamp: int  # noqa: N815
    contentType: str  # noqa: N815


class B2FileInformationNotAvailableException(Exception):  # noqa: N818
    ...


class BackblazeB2StorageConstructorArgs(PossibleB2StorageOptions):
    opts: NotRequired[PossibleB2StorageOptions]


@deconstructible
class BackblazeB2Storage(Storage):
    """Storage class which fulfills the Django Storage contract through b2 apis"""

    def __init__(self, **kwargs: Unpack[BackblazeB2StorageConstructorArgs]):
        constructor_options = self._extract_constructor_options(
            cast(PossibleB2StorageOptions, kwargs), kwargs.pop("opts", PossibleB2StorageOptions())
        )
        django_settings_options = self._get_options_from_django_settings()
        opts = _merge(source=constructor_options, into=django_settings_options)  # type: ignore[arg-type]
        logger.debug(
            f"Initializing {self.__class__.__name__} with options "
            + str({**opts, "application_key_id": "<redacted>", "application_key": "<redacted>"})
        )

        self._bucket_name = opts["bucket"]
        self._default_file_metadata = opts["default_file_info"]
        self._forbid_file_property_caching = opts["forbid_file_property_caching"]
        self._authInfo = dict(
            [(k, v) for k, v in opts.items() if k in ["realm", "application_key_id", "application_key"]]
        )
        self._allow_file_overwrites = opts["allow_file_overwrites"]

        self._get_account_info = self._create_account_info_callable(opts)

        logger.info(f"{self.__class__.__name__} instantiated to use bucket {self._bucket_name}")
        if opts["authorize_on_init"]:
            logger.debug(f"{self.__class__.__name__} authorizing")
            self.b2_api
            if opts["validate_on_init"]:
                self._get_or_create_bucket(opts["non_existent_bucket_details"])

    def _extract_constructor_options(
        self, constructor_kwargs: PossibleB2StorageOptions, opts_args: PossibleB2StorageOptions
    ) -> PossibleB2StorageOptions:
        if constructor_kwargs and opts_args:
            raise ImproperlyConfigured("Can only specify opts or keyword args, not both!")
        options = constructor_kwargs or opts_args
        self._validate_options(options, from_constructor=True)
        return options

    def _get_options_from_django_settings(self) -> BackblazeB2StorageOptions:
        """Setting terminology taken from:
        https://b2-sdk-python.readthedocs.io/en/master/glossary.html#term-application-key-ID
        kwargOpts available for subclasses
        """
        from django.conf import settings

        if not hasattr(settings, "BACKBLAZE_CONFIG"):
            raise ImproperlyConfigured("add BACKBLAZE_CONFIG dict to django settings")
        if "application_key_id" not in settings.BACKBLAZE_CONFIG or "application_key" not in settings.BACKBLAZE_CONFIG:
            raise ImproperlyConfigured(
                "At minimum BACKBLAZE_CONFIG must contain auth 'application_key' and 'application_key_id'"
                f"\nfound: {settings.BACKBLAZE_CONFIG}"
            )
        self._validate_options(cast(PossibleB2StorageOptions, settings.BACKBLAZE_CONFIG))
        opts = get_default_b2_storage_options()
        opts.update(settings.BACKBLAZE_CONFIG)  # type: ignore[typeddict-item]
        return opts

    def _validate_options(self, options: PossibleB2StorageOptions, from_constructor: bool = False) -> None:
        unrecognized_options = [k for k in options.keys() if k not in get_default_b2_storage_options().keys()]
        if unrecognized_options:
            raise ImproperlyConfigured(f"Unrecognized options: {unrecognized_options}")

    def _create_account_info_callable(self, opts: BackblazeB2StorageOptions) -> Callable[[], AbstractAccountInfo]:
        if (
            not isinstance(opts["account_info"], dict)
            or "type" not in opts["account_info"]
            or opts["account_info"]["type"] not in ["memory", "sqlite", "django-cache"]
        ):
            raise ImproperlyConfigured(
                (f"'account_info' property must be a dict with type found in options.py, was {opts['account_info']}")
            )
        if opts["account_info"]["type"] == "django-cache":
            logger.debug(f"{self.__class__.__name__} will use {DjangoCacheAccountInfo.__name__}")
            return lambda: DjangoCacheAccountInfo(
                cache_name=cast(DjangoCacheAccountInfoConfig, opts["account_info"]).get("cache", "django-backblaze-b2")
            )
        elif opts["account_info"]["type"] == "memory":
            logger.debug(f"{self.__class__.__name__} will use {InMemoryAccountInfo.__name__}")
            return lambda: InMemoryAccountInfo()
        elif opts["account_info"]["type"] == "sqlite":
            logger.debug(f"{self.__class__.__name__} will use {SqliteAccountInfo.__name__}")
            return lambda: SqliteAccountInfo(
                file_name=cast(SqliteAccountInfoConfig, opts["account_info"])["database_path"]
            )
        raise ImproperlyConfigured()

    @property
    def b2_api(self) -> B2Api:
        if not hasattr(self, "_b2_api"):
            self._account_info = self._get_account_info()
            self._b2_api = B2Api(account_info=self._account_info, cache=AuthInfoCache(self._account_info))
            self._b2_api.authorize_account(**self._authInfo)
        return self._b2_api

    @property
    def bucket(self) -> Bucket:
        if not hasattr(self, "_bucket"):
            self._get_or_create_bucket()
        return self._bucket

    def _get_or_create_bucket(self, new_bucket_details=None) -> None:
        try:
            self._bucket = self.b2_api.get_bucket_by_name(self._bucket_name)
        except NonExistentBucket as e:
            if new_bucket_details is not None:
                logger.debug(f"Bucket {self._bucket_name} not found. Creating with details: {new_bucket_details}")
                if "bucket_type" not in new_bucket_details:
                    new_bucket_details["bucket_type"] = "allPrivate"
                self._bucket = self.b2_api.create_bucket(name=self._bucket_name, **new_bucket_details)
            else:
                raise e
        logger.debug(f"Connected to bucket {self._bucket.as_dict()}")

    def _refresh_bucket(self) -> Bucket:
        if self._bucket:
            return self._bucket.get_fresh_state()
        return self.bucket

    def _open(self, name: str, mode: str) -> File:
        return B2File(
            name=name,
            bucket=self.bucket,
            file_metadata=self._default_file_metadata,
            mode=mode,
            size_provider=self.size,
        )

    def _save(self, name: str, content: IO[Any]) -> str:
        """
        Save and retrieve the filename.
        If the file exists it will make another version of that file.
        """
        return B2File(
            name=name,
            bucket=self.bucket,
            file_metadata=self._default_file_metadata,
            mode="w",
            size_provider=self.size,
        ).save_and_retrieve_file(content)

    def path(self, name: str) -> str:
        return name

    def delete(self, name: str) -> None:
        file_info = self._file_info(name)
        if file_info:
            logger.debug(f"Deleting file {name} id=({file_info['fileId']})")
            self.b2_api.delete_file_version(file_id=file_info["fileId"], file_name=name)
            if self._cache:
                self._cache.delete(self._file_cache_key(name))
        else:
            logger.debug("Not found")

    def _file_info(self, name: str) -> Optional[_SdkFileInfoDict]:
        try:
            if self._cache:
                cache_key = self._file_cache_key(name)
                timeout_in_seconds = 60

                def load_info():
                    logger.debug(f"file info cache miss for {name}")
                    return self.bucket.get_file_info_by_name(name).as_dict()

                return self._cache.get_or_set(key=cache_key, default=load_info, timeout=timeout_in_seconds)
            return self.bucket.get_file_info_by_name(name).as_dict()
        except FileNotPresent:
            return None

    def _file_cache_key(self, name: str) -> str:
        return hash(f"{self.bucket.name}__{name}".encode()).hexdigest()

    @property
    def _cache(self) -> Optional[BaseCache]:
        if (
            not self._forbid_file_property_caching
            and self.b2_api  # force init
            and self._account_info
            and isinstance(self._account_info, DjangoCacheAccountInfo)
        ):
            return self._account_info.cache
        return None

    def exists(self, name: str) -> bool:
        return bool(self._file_info(name))

    def size(self, name: str) -> int:
        file_info = self._file_info(name)
        return file_info.get("size", 0) if file_info else 0

    def url(self, name: Optional[str]) -> str:
        if not name:
            raise Exception("Name must be defined")
        return self._get_file_url(name)

    def _get_file_url(self, name: str) -> str:
        return self.get_backblaze_url(name)

    def get_backblaze_url(self, filename: str) -> str:
        return self.b2_api.get_download_url_for_file_name(bucket_name=self._bucket_name, file_name=filename)

    def get_available_name(self, name: str, max_length: Optional[int] = None) -> str:
        if self._allow_file_overwrites:
            return name
        return super().get_available_name(name, max_length)

    def listdir(self, path: str) -> Tuple[List[str], List[str]]:
        """
        List the contents of the specified path. Return a 2-tuple of lists:
        the first item being directories, the second item being files.
        """
        raise NotImplementedError("subclasses of Storage must provide a listdir() method")

    def get_accessed_time(self, name: str) -> datetime:
        """
        Return the last accessed time (as a datetime) of the file specified by
        name. The datetime will be timezone-aware if USE_TZ=True.
        """
        raise NotImplementedError("subclasses of Storage must provide a get_accessed_time() method")

    def get_created_time(self, name: str) -> datetime:
        """
        Return the creation time (as a datetime) of the file specified by name.
        The datetime will be timezone-aware if USE_TZ=True.
        """
        from datetime import timezone

        from django.conf import settings

        file_info = self._file_info(name)
        try:
            if file_info and float(file_info.get("uploadTimestamp", 0)) > 0:
                timestamp = float(file_info["uploadTimestamp"]) / 1000.0
                if settings.USE_TZ:
                    return datetime.fromtimestamp(timestamp, timezone.utc)
                return datetime.fromtimestamp(timestamp)
        except ValueError as e:
            raise B2FileInformationNotAvailableException(f"'uploadTimestamp' from API not valid for {name}: {e}")
        raise B2FileInformationNotAvailableException(f"'uploadTimestamp' not available for {name}")

    def get_modified_time(self, name: str) -> datetime:
        """
        Return the last modified time (as a datetime) of the file specified by
        name. The datetime will be timezone-aware if USE_TZ=True.
        """
        return self.get_created_time(name)


T = TypeVar("T")


def _merge(source: Dict, into: T, path=None) -> T:
    """merges b into a
    https://stackoverflow.com/a/7205107/11076240
    """
    target = cast(dict, into)  # easier to read 'target' within function body
    merged = target.copy()
    if path is None:
        path = []
    for key in source:
        if key in target:
            printable_path = ".".join(path + [str(key)])
            if isinstance(target[key], dict) and isinstance(source[key], dict):
                merged[key] = _merge(source=source[key], into=target[key], path=path + [str(key)])
            elif target[key] != source[key]:
                logger.debug(f"Overriding setting {printable_path} '{target[key]}' with value '{source[key]}'")
                merged[key] = source[key]
        else:
            merged[key] = source[key]
    return cast(T, merged)



from django.urls import path

from django_backblaze_b2 import views

app_name = "django_b2_storage"
urlpatterns = [
    path("b2/<path:filename>", views.download_public_file, name="b2-public"),
    path("b2l/<path:filename>", views.download_logged_in_file, name="b2-logged-in"),
    path("b2s/<path:filename>", views.download_staff_file, name="b2-staff"),
]



from io import BytesIO
from logging import getLogger
from typing import IO, Any, Callable, Dict, Optional, Union

from b2sdk.v2 import Bucket
from django.core.files.base import File

logger = getLogger("django-backblaze-b2")


class B2File(File):
    """Read/Write as lazy as possible"""

    def __init__(
        self,
        name: str,
        bucket: Bucket,
        size_provider: Callable[[str], int],
        file_metadata: Dict[str, Any],
        mode: str,
    ):
        self.name: str = name
        self._bucket: Bucket = bucket
        self._size_provider = size_provider
        self._file_metadata = file_metadata
        self._mode: str = mode
        self._has_unwritted_data: bool = False
        self._contents: Optional[IO] = None

    @property
    def file(self) -> Union[IO[Any], None]:
        if self._contents is None:
            self._contents = self._read_file_contents()
        return self._contents

    @file.setter
    def file(self, value: IO[Any]) -> None:
        self._contents = value

    def _read_file_contents(self) -> BytesIO:
        currently_downloading_file = self._bucket.download_file_by_name(file_name=self.name)
        bytes_io = BytesIO()
        currently_downloading_file.save(bytes_io)
        contents = BytesIO(bytes_io.getvalue())
        bytes_io.close()
        return contents

    @property
    def size(self) -> int:
        if not hasattr(self, "_size"):
            self._size = self._size_provider(self.name)
        return self._size

    def read(self, num_bytes: Optional[int] = None) -> bytes:
        return self.file.read(num_bytes if isinstance(num_bytes, int) else -1) if self.file else bytes()

    def write(self, content) -> int:
        if "w" not in self._mode:
            raise AttributeError("File was not opened for write access.")
        self.file = BytesIO(content)

        self._has_unwritted_data = True
        return len(content)

    def close(self) -> None:
        if self.file:
            if self._has_unwritted_data:
                self.save_and_retrieve_file(self.file)
            self.file.close()

    def save_and_retrieve_file(self, content: IO[Any]) -> str:
        """
        Save and retrieve the filename.
        If the file exists it will make another version of that file.
        """
        logger.debug(f"Saving {self.name} to b2 bucket ({self._bucket.get_id()})")
        self._bucket.upload_bytes(
            data_bytes=content.read(),
            file_name=self.name,
            file_info=self._file_metadata,
        )
        self._has_unwritted_data = False
        return self.name



import django_backblaze_b2.urls as urls

from .storage import BackblazeB2Storage
from .storages import LoggedInStorage, PublicStorage, StaffStorage

__version__ = "6.0.0"
__all__ = ["PublicStorage", "LoggedInStorage", "StaffStorage", "BackblazeB2Storage", "urls"]



from __future__ import annotations

from typing import Any, Dict, Optional, Union

from typing_extensions import Literal, TypedDict


class PossibleB2StorageOptions(TypedDict, total=False):
    realm: str  # default "production"
    application_key_id: str
    application_key: str
    bucket: str
    authorize_on_init: bool
    validate_on_init: bool
    allow_file_overwrites: bool
    account_info: Optional[Union[DjangoCacheAccountInfoConfig, InMemoryAccountInfoConfig, SqliteAccountInfoConfig]]
    forbid_file_property_caching: bool
    specific_bucket_names: ProxiedBucketNames
    cdn_config: Optional[CDNConfig]
    # see: https://b2-sdk-python.readthedocs.io/en/master/api/api.html#b2sdk.v1.B2Api.create_bucket
    non_existent_bucket_details: Optional[Dict[str, Union[str, Dict[str, Any]]]]
    default_file_info: Dict[str, Any]


class BackblazeB2StorageOptions(TypedDict):
    realm: str
    application_key_id: str
    application_key: str
    bucket: str
    authorize_on_init: bool
    validate_on_init: bool
    allow_file_overwrites: bool
    account_info: Optional[Union[DjangoCacheAccountInfoConfig, InMemoryAccountInfoConfig, SqliteAccountInfoConfig]]
    forbid_file_property_caching: bool
    specific_bucket_names: ProxiedBucketNames
    cdn_config: Optional[CDNConfig]
    non_existent_bucket_details: Optional[Dict[str, Union[str, Dict[str, Any]]]]
    default_file_info: Dict[str, Any]


def get_default_b2_storage_options() -> BackblazeB2StorageOptions:
    return {
        "realm": "production",
        "application_key_id": "you must set this value yourself",
        "application_key": "you must set this value yourself",
        "bucket": "django",
        "authorize_on_init": True,
        "validate_on_init": True,
        "allow_file_overwrites": False,
        "account_info": {"type": "django-cache", "cache": "django-backblaze-b2"},
        "forbid_file_property_caching": False,
        "specific_bucket_names": {"public": None, "logged_in": None, "staff": None},
        "cdn_config": None,
        "non_existent_bucket_details": None,
        "default_file_info": {},
    }


class ProxiedBucketNames(TypedDict, total=False):
    public: Optional[str]
    logged_in: Optional[str]
    staff: Optional[str]


class DjangoCacheAccountInfoConfig(TypedDict):
    type: Literal["django-cache"]
    cache: str


class InMemoryAccountInfoConfig(TypedDict):
    type: Literal["memory"]


class SqliteAccountInfoConfig(TypedDict):
    type: Literal["sqlite"]
    database_path: str


class CDNConfig(TypedDict):
    base_url: str
    include_bucket_url_segments: bool



import logging
import threading
from functools import wraps
from hashlib import sha3_224 as hash
from typing import Iterable, List, Optional, Tuple

from b2sdk.v2 import UrlPoolAccountInfo
from b2sdk.v2.exception import MissingAccountData
from django.core.cache import InvalidCacheBackendError, caches
from django.core.exceptions import ImproperlyConfigured

logger = logging.getLogger("django-backblaze-b2")


class StoredBucketInfo:
    name: str
    id_: str


def _handle_result_is_none(item_name=None):
    """
    Raise MissingAccountData if function's result is None.
    adapted from https://github.com/Backblaze/b2-sdk-python/blob/v1.5.0/b2sdk/account_info/in_memory.py
    """

    def wrapper_factory(function):
        @wraps(function)
        def getter_function(self, *args, **kwargs):
            assert function.__name__.startswith("get_")
            result = function(self, *args, **kwargs)
            if result is None:
                self.cache.clear()
                raise MissingAccountData(
                    f"Token refresh required to determine value of '{item_name or function.__name__[len('get_') :]}'",
                )
            return result

        return getter_function

    return wrapper_factory


class DjangoCacheAccountInfo(UrlPoolAccountInfo):
    """
    Store account information in django's cache: https://docs.djangoproject.com/en/3.1/topics/cache

    Threadsafe only in the context of its own runtime,
    i.e. it's possible (unlikely though, only in a highly concurrent scenario) that a
    wrapper of Django could launch 2+ different processes, which would not share threads,
    and within the 'locked' blocks, mutation of values between cache accesses
    """

    def __init__(self, cache_name: str):
        logger.debug(f"Initializing {self.__class__.__name__} with cache '{cache_name}'")
        self._cache_name = cache_name
        self._cache_lock = threading.Lock()
        try:
            self.cache = caches[cache_name]
            self.cache.set("bucket_names", [])
        except InvalidCacheBackendError:
            logger.exception("Cache assignment failed")
            from django.conf import settings

            help_message = (
                (
                    ". "
                    "The default 'account_info' option of this library is with a django cache"
                    " by the name of 'django-backblaze-b2'"
                )
                if "account_info" not in settings.BACKBLAZE_CONFIG
                else ""
            )

            raise ImproperlyConfigured(
                f"Expected to find a cache with name '{cache_name}' as per options" + help_message
            )
        super(DjangoCacheAccountInfo, self).__init__()

    def clear(self):
        """
        Remove all info about accounts and buckets.
        """
        self.cache.clear()
        self.cache.set("bucket_names", [])

    def _set_auth_data(
        self,
        account_id,
        auth_token,
        api_url,
        download_url,
        recommended_part_size,
        absolute_minimum_part_size,
        application_key,
        realm,
        s3_api_url,
        allowed,
        application_key_id,
    ):
        self.cache.set(
            "cached_account_info",
            {
                "account_id": account_id,
                "auth_token": auth_token,
                "api_url": api_url,
                "download_url": download_url,
                "recommended_part_size": recommended_part_size,
                "absolute_minimum_part_size": absolute_minimum_part_size,
                "application_key": application_key,
                "realm": realm,
                "s3_api_url": s3_api_url,
                "allowed": allowed,
                "application_key_id": application_key_id,
            },
            timeout=None,
        )

    def _cached_info(self):
        return self.cache.get("cached_account_info", default={})

    @_handle_result_is_none()
    def get_application_key(self):
        return self._cached_info().get("application_key")

    @_handle_result_is_none()
    def get_application_key_id(self):
        return self._cached_info().get("application_key_id")

    @_handle_result_is_none()
    def get_account_id(self):
        return self._cached_info().get("account_id")

    @_handle_result_is_none()
    def get_api_url(self):
        return self._cached_info().get("api_url")

    @_handle_result_is_none("auth_token")
    def get_account_auth_token(self):
        """Named different from cached value"""
        return self._cached_info().get("auth_token")

    @_handle_result_is_none()
    def get_download_url(self):
        return self._cached_info().get("download_url")

    @_handle_result_is_none()
    def get_realm(self):
        return self._cached_info().get("realm")

    @_handle_result_is_none()
    def get_absolute_minimum_part_size(self):
        return self._cached_info().get("absolute_minimum_part_size")

    @_handle_result_is_none()
    def get_recommended_part_size(self):
        return self._cached_info().get("recommended_part_size")

    @_handle_result_is_none()
    def get_allowed(self):
        return self._cached_info().get("allowed")

    def get_s3_api_url(self):
        return self._cached_info().get("s3_api_url") or ""

    def get_bucket_id_or_none_from_bucket_name(self, bucket_name: str) -> Optional[str]:
        try:
            return self.cache.get(_bucket_cachekey(bucket_name))
        except KeyError as e:
            logger.debug(f"cache miss {bucket_name}: {e}")
            return None

    def get_bucket_name_or_none_from_bucket_id(self, bucket_id: str) -> Optional[str]:
        try:
            self._cache_lock.acquire()
            for bucket_name in self.cache.get("bucket_names", []):
                cached_id = self.cache.get(_bucket_cachekey(bucket_name))
                if cached_id and cached_id == bucket_id:
                    return bucket_name
            logger.debug(f"cache miss {bucket_id}")
        except KeyError as e:
            logger.debug(f"cache miss {bucket_id}: {e}")
        finally:
            self._cache_lock.release()
        return None

    def refresh_entire_bucket_name_cache(self, name_id_iterable: Iterable[Tuple[str, str]]):
        with self._cache_lock:
            new_bucket_names = set()
            for bucket_name, bucket_id in name_id_iterable:
                self.cache.set(_bucket_cachekey(bucket_name), bucket_id)
                new_bucket_names.add(bucket_name)

            buckets_to_remove = [n for n in self.cache.get("bucket_names", []) if n not in new_bucket_names]
            for bucket_name in buckets_to_remove:
                self.cache.delete(_bucket_cachekey(bucket_name))

            self.cache.set("bucket_names", list(new_bucket_names))

    def save_bucket(self, bucket: StoredBucketInfo):
        with self._cache_lock:
            self.cache.set(_bucket_cachekey(bucket.name), bucket.id_)
            self.cache.set("bucket_names", self.cache.get("bucket_names", []) + [bucket.name])

    def remove_bucket_name(self, bucket_name):
        with self._cache_lock:
            self.cache.set("bucket_names", [n for n in self.cache.get("bucket_names", []) if n != bucket_name])
            self.cache.delete(_bucket_cachekey(bucket_name))

    def list_bucket_names_ids(self) -> List[Tuple[str, str]]:
        tuples = []
        with self._cache_lock:
            for bucket_name in self.cache.get("bucket_names", []):
                bucket_id = self.cache.get(_bucket_cachekey(bucket_name))
                tuples.append((bucket_name, bucket_id))
        return tuples

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}{{cache_name={self._cache_name}, cache={self.cache}}}"


def _bucket_cachekey(bucket_name: str) -> str:
    return hash(f"bucket-name__{bucket_name}".encode()).hexdigest()



import logging
import mimetypes
from typing import Union

from b2sdk.v1.exception import FileNotPresent
from django.http import FileResponse, HttpRequest, HttpResponse, HttpResponseNotFound
from django.utils.translation import gettext_lazy as _
from django.views.decorators.clickjacking import xframe_options_sameorigin

from django_backblaze_b2 import storage, storages

from ._decorators import _requires_login

logger = logging.getLogger("django-backblaze-b2")


@xframe_options_sameorigin
def download_public_file(request: HttpRequest, filename: str) -> Union[HttpResponse, FileResponse]:
    """Serves the specified 'filename' without validating any authentication"""
    return _download_file_from_storage(storages.PublicStorage(), filename)


@_requires_login()
@xframe_options_sameorigin
def download_logged_in_file(request: HttpRequest, filename: str) -> Union[HttpResponse, FileResponse]:
    """Serves the specified 'filename' validating the user is logged in"""
    return _download_file_from_storage(storages.LoggedInStorage(), filename)


@_requires_login(requires_staff=True)
@xframe_options_sameorigin
def download_staff_file(request: HttpRequest, filename: str) -> Union[HttpResponse, FileResponse]:
    """Serves the specified 'filename' validating the user is logged in and a staff user"""
    return _download_file_from_storage(storages.StaffStorage(), filename)


def _download_file_from_storage(
    storage: storage.BackblazeB2Storage, filename: str
) -> Union[HttpResponse, FileResponse]:
    if logger.isEnabledFor(logging.DEBUG):
        try:
            logger.debug(f"Downloding file from {storage.get_backblaze_url(filename)}")
        except Exception:
            logger.exception(f"Debug log failed. Could not retrive b2 file url for {filename}")

    try:
        if storage.exists(filename):
            content_type, _encoding = mimetypes.guess_type(filename)
            return FileResponse(storage.open(filename, "r"), content_type=content_type)
    except (FileNotFoundError, FileNotPresent):
        logging.exception("Opening backblaze file failed")

    return HttpResponseNotFound(_("Could not find file") + f": {filename}")



from functools import wraps

from django.http import HttpRequest, HttpResponse
from django.utils.translation import gettext_lazy as _


def _requires_login(requires_staff=False):
    """
    Decorator for views that checks that the user passes the given test,
    redirecting to the log-in page if necessary. The test should be a callable
    that takes the user object and returns True if the user passes.
    """

    def decorator(view_func):
        @wraps(view_func)
        def _wrapped_view(request: HttpRequest, *args, **kwargs):
            from django.contrib.auth.views import redirect_to_login  # delay import in case custom user model

            if request.user.is_anonymous:
                return redirect_to_login(request.get_full_path())
            if not request.user.is_active or (requires_staff and not request.user.is_staff):
                return HttpResponse(_("Unauthorized"), status=401)
            return view_func(request, *args, **kwargs)

        return _wrapped_view

    return decorator
